------------------------------------------------------------------------
--  GraphicCtrl.vhd -- Pabi Waledi game Controller
------------------------------------------------------------------------
--  Author : Mircea Dabacan
--           Copyright 2004 Digilent, Inc.
------------------------------------------------------------------------
-- Software version: Xilinx ISE 6.2.03i
--                   WebPack
------------------------------------------------------------------------
-- This is the source file for the GraphicCtrl component, 
-- used in the "Tennis" Digilent Reference Project.

-- This file contains the design for a Pabi Waledi game Controller to be 
-- included in a user project and downloaded to a Digilent FPGA System 
-- board. This variant includes LCD messages displayed during the game.

-- The project uses 6 buttons:
-- Reset       - Btn1 on the main board
--                  DIO2 or       DIO1      DIO4      Pegasus or
--                  DIO5                              Spartan 3
-- Start Game    - button B      btn 3      btn 3
-- Hit Right     - button E      btn 5      btn 1      btn 0
-- Hit Left      - button 1      btn 1      btn 5      btn 3
-- Turbo Right   - button D      btn 4      btn 2      btn 1
-- Turbo Left    - button 0      btn 2      btn 4      btn 2

-- The play field is symbolized by the LED string. The ball is the ON  
-- LED while all the other LEDs are OFF.
-- When the "Start Game" button is pressed, a random algorithm decides 
-- which player serves, placing the ball (steady) on his extreme. 
-- The server player hits the ball, pressing the corresponding button. 
-- The ball moves toward the other player. The player must hit the ball
-- (press the corresponding "hit" button) when the ball is on his 
-- extreme position (last LED of the string at his end). Attempting to
-- hit the ball earlier (i.e. when the ball is on the second to extreme
-- position) as well as failing to hit the ball when on the extreme 
-- position, is a mistake and leads to losing the point. 
-- If the ball is hit correctly, it reverses the move direction, the 
-- other player must hit it when on his extreme.
-- The speed of the ball can be controlled by the player who had hit it
-- last: if his "turbo" button is released, the ball moves around 6 
-- positions per second. While he keeps his "turbo" button pressed, the
-- ball's speed doubles.
-- The extreme positions of the 7-segment LED display keep the score for
-- each player. A game ends when a player accumulates 9 points. 
-- Pressing the "Start Game" button starts a new game.
-- The server is again randomly decided.
-- If an LCD display is available on the board (combination) it displays 
-- messages showing the game status.

------------------------------------------------------------------------
--  Port definitions
------------------------------------------------------------------------
-- Pabi_Valedi_FSM interfaces either to a bus controller (when using  
-- Digilab DIO2 or DIO5  boards, or directly to LEDs and a 
-- seven segment display controller, for any other board (combination). 
-- When an LCD display is available, it is accessed through a message 
-- controller (i.e. Lcd_Mess_Ctrl). For details about using these 
-- interface components, see the Digilent Component Library.

-- The used signals are:
-- general signals
   -- ck, RESET: in std_logic; 
-- keyboard signals
   -- hit_left,hit_right,start_game,turbo_left,turbo_right:in std_logic;
-- LEDs
   -- led_state: inout std_logic_vector (7 downto 0);      
         -- feedback requires inout
-- Score - interface to dbIOxCtrl
   -- score_left, score_right: inout integer range 0 to 9;
         -- feedback requires inout
-- Interface to LcdMessCtrl signals
   -- LcdCfgString : out config_string_type;   
         -- LCD configuration string
   -- LcdMessage: out string;      -- LCD text message
   -- lcd_mes_req: inout std_logic; -- request to send a message to LCD
   -- lcd_mes_ack: in std_logic); -- dbIOxCtrl begun sending the message

-- The LcdCfgString is an array of std_logic_vector(7 downto 0);
-- This type is not recognized by default, so it needs to be declared.
-- To be used in the entity, the type must be declared somewhere before.
-- The single way to do that is using a package, which is lcd_pkg.vhd 

------------------------------------------------------------------------
--  Internal signals
------------------------------------------------------------------------
--   sreg_PLAY: type_sreg_PLAY; 
         -- state register (enumerated type, see the source code)
--   random_serv: std_logic;            -- registered (0 = left player)
--   ck_div_cnt: std_logic_vector(1 downto 0); -- divider by 2 or 2^2
--   play_ck: std_logic;                -- 6Hz or 12Hz clock for the FSM

-- The constants in the architecture declaration segment define
-- configuration and text messages for the LCD. 

------------------------------------------------------------------------
--  Behavioral description
------------------------------------------------------------------------
-- "random" process generates a random value (the value of a very fast 
-- changing bit sampled when the user presses the "Start Game button". 
-- This value is further used to decide which player begins a new game.

-- The "Clock_divider" process generates the clock for the game FSM. 
-- Its' frequency is responsible for  the apparent speed of the ball or 
-- for the speed of the dynamic light at the end of each ball. 
-- Higher clock frequency is generated when the player who had hit the
-- ball last press "turbo" button, as well as for the "dynamic light".

-- A BUFG component is instantiated to drive the play_ck.

-- The FSM process implements the game FSM:
-- For details see the comments in the source code and the 
-- "Tennis Design Documents" file.

------------------------------------------------------------------------
-- Revision History:
--   12/21/2004(MirceaD): created
--   03/24/2005(MirceaD): formatted
------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
library work;
use work.LCD_pkg.all;

entity GraphicCtrl is
   port (ck, RESET: in std_logic; -- general signals
   -- keyboard signals
   -- Interface to LcdMessCtrl signals
      LcdCfgString : out config_string_type(1 to 6);   
          -- LCD configuration string   
      LcdMessage : out string(1 to 41);             -- LCD text message
      lcd_mes_req : inout std_logic; -- request to send a message to LCD  
      lcd_mes_ack : in std_logic);-- dbIOxCtrl begun sending the message
end;

architecture behavioral of GraphicCtrl is

constant LcdInitString : config_string_type(1 to 6):=
        (x"40", -- Set Y (column) address to 0
         x"B8", -- Set X (row) address to 0
         x"C0", -- Set display line to 0
         x"3F", -- Set Display ON
         x"00");-- Null (never sent)
-- The following constants define the messages for the LCD. The last 
-- character in each message, "|", is recognized by the Lcd_Mess_Ctrl 
-- component as the message "end mark". 
constant Enjoy_tennis    : string := 
                           "Enjoy tennis! Copyright Digilent, 2004. |";
constant Press_B_to_start: string := "Press B to start|";
constant Left_mistake    : string := "Left mistake|";
constant Right_mistake   : string := "Right mistake|";
constant Left_player_won : string := "Left player won|";
constant Right_player_won: string := "Right player won|";
constant Left_service    : string := "Left service|";
constant Right_service   : string := "Right service|";

component BUFG      -- bufg declared for play_ck
   port (I: in std_logic; O: out std_logic);
end component;   

-- Actual architecture begin:
begin

random: process (start_game)
   begin
-- ck_div_cnt(0) counts at 5Mhz/2^21 = 23.84Hz and it is asynchronous to
-- dbIO2Ctrl (which uses a 50MHz/16 = 3.125MHz). Aat the time when key 
-- "Start Game" is transmitted, the value of ck_div_cnt(0)  is random!
   if start_game'event and start_game = '1' then
      random_serv <= ck_div_cnt(0);
   end if;
end process;

clock_divider: process (ck)
   begin
   if ck'event and ck='1' then
      if ((sreg_PLAY = go_left and turbo_right='1') or      
               -- right player did hit the ball and presses "turbo"
          (sreg_PLAY = go_right and turbo_left='1') or      
               -- left player did hit the ball and presses "turbo"
          (sreg_PLAY = lost_left) or (sreg_PLAY = lost_right)) then 
              -- the dynamic light after ending a ball
          ck_div_cnt <= ck_div_cnt + "10"; 
              -- dividing 24Hz by 2 results in ~12Hz.
      else
          ck_div_cnt <= ck_div_cnt + "01";   
              -- dividing 24Hz by 2^2 results in ~6Hz.
      end if;
   end if;
end process;

--play_ck<=ck_div_cnt(1);   -- replaced by the BUFG instantiation below
play_ck_buf: BUFG port map (I => ck_div_cnt(1), O => play_ck);      
              -- bufg instantiated for play_ck

-- The game FSM
FSM: process (play_ck, RESET)
begin
  if ( RESET='1' ) then
      led_state <= "00000000";
      score_left <= 0;
      score_right <= 0;
      LCD_mes_req <= '0';
      sreg_PLAY <= m_Start_up;
   elsif play_ck='1' and play_ck'event then
      case sreg_PLAY is
         when m_Start_up => 
              -- Send message:"Enjoy tennis! Copyright Digilent, 2004. "
            if LCD_mes_req = '0' and LCD_mes_ack ='0' then 
                                             -- no message pending
               LCD_mes_req <= '1';
               LcdCfgString(1 to LcdInitString'length) <= 
                                                        LcdInitString;
               LcdMessage(1 to Enjoy_tennis'length) <= Enjoy_tennis;                        
--             sreg_PLAY <= m_Start_up;      -- Keep old state (default)
            elsif LCD_mes_req = '1' and LCD_mes_ack ='1' then   
                                    -- the last message was acknowledged
               LCD_mes_req <= '0';           -- remove the request
               sreg_PLAY<=m_Press_B;
--          else
--             sreg_PLAY <= m_Start_up;      -- Keep old state (default)
            end if;

         when m_Press_B =>      -- Send messages: "Press B to start" 
            if LCD_mes_req = '0' and LCD_mes_ack ='0' then 
                                             -- no message pending
               LCD_mes_req <= '1';
               LcdCfgString(1 to LCD_lr_string'length) <= 
                                                          LCD_lr_string;
               LcdMessage( 1 to Press_B_to_start'length) <= 
                                                       Press_B_to_start;
--             sreg_PLAY <= m_Press_B;       -- Keep old state (default)
            elsif LCD_mes_req = '1' and LCD_mes_ack ='1' then   
                                    -- the last message was acknowledged
               LCD_mes_req <= '0';           -- remove the request
               sreg_PLAY<=Shift_mess;
--          else
--             sreg_PLAY <= m_Press_B;       -- Keep old state (default)
            end if;

         when Shift_mess =>   
                 -- Shift the message, until key "Start Game" is pressed
            if LCD_mes_req = '0' and LCD_mes_ack ='0' then 
                                             -- no message pending
               if (start_game = '1') then
                                   -- the "Start game" button is pressed
                  score_right <= 0;          -- initialize the scores
                  score_left <= 0;
                  if random_serv = '0' then-- decide which player serves
                     led_state <= "10000000";     
                                        -- set the initial ball position
                     sreg_PLAY <= m_start_left;
                  else
                     led_state <= "00000001";
                                        -- set the initial ball position
                     sreg_PLAY <= m_start_right;
                  end if;
               else
                  LCD_mes_req <= '1';
                  LcdCfgString(1 to LCD_shift_right_string'length) 
                                              <= LCD_shift_right_string;
--                LcdMessage(1) <= '|';         -- NULL text message
                  LcdMessage(1 to 2) <= "| ";   
       -- should be the line above, but that doesn't work (no idea why?)
--                sreg_PLAY <= Shift_mess;   -- Keep old state (default)
               end if;
            elsif LCD_mes_req = '1' and LCD_mes_ack ='1' then   
                                    -- the last message was acknowledged
               LCD_mes_req <= '0';           -- remove the request
--             sreg_PLAY <= Shift_mess;      -- Keep old state (default)
--          else
--             sreg_PLAY <= Shift_mess;      -- Keep old state (default)
            end if;

         when m_start_left =>            -- Send message: "Left service"
            if LCD_mes_req = '0' and LCD_mes_ack ='0' then 
                                             -- no message pending
               LCD_mes_req <= '1';
               LcdCfgString(1 to LcdInitString'length) <= 
                                                        LcdInitString;
               LcdMessage(1 to Left_service'length) <= Left_service;
--             sreg_PLAY <= m_start_left;    -- Keep old state (default)
            elsif LCD_mes_req = '1' and LCD_mes_ack ='1' then   
                                    -- the last message was acknowledged
               LCD_mes_req <= '0';                 -- remove the request
               sreg_PLAY <= start_left;
--          else
--             sreg_PLAY <= m_start_left;    -- Keep old state (default)
            end if;

         when m_start_right =>      -- Send message: "Right service"
            if LCD_mes_req = '0' and LCD_mes_ack ='0' then 
                                    -- no message pending
               LCD_mes_req <= '1';
               LcdCfgString(1 to LcdInitString'length) <= 
                                                        LcdInitString;
               LcdMessage(1 to Right_service'length) <= Right_service;
--             sreg_PLAY <= m_start_right;   -- Keep old state (default)
            elsif LCD_mes_req = '1' and LCD_mes_ack ='1' then   
                                    -- the last message was acknowledged
               LCD_mes_req <= '0';           -- remove the request
               sreg_PLAY<=start_right;
--          else
--             sreg_PLAY <= m_start_right;   -- Keep old state (default)
            end if;

         when start_left =>
            if ( hit_left='1' ) then         -- the left player serves
               led_state<='0'&led_state(7 downto 1);-- shift LEDs right
               sreg_PLAY<=go_right;
--          else
--             sreg_PLAY<=start_left;        -- Keep old state (default)
            end if;

         when start_right =>
            if ( hit_right='1' ) then        -- the right player serves
               led_state <=led_state(6 downto 0)&'0';-- shift LEDs left
               sreg_PLAY<=go_left;
--          else
--             sreg_PLAY<=start_right;       -- Keep old state (default)
            end if;

         when go_right =>
            if ( led_state(0)='1' ) and ( hit_right='1' ) then   
                                             -- correct return
               led_state <= led_state(6 downto 0)&'0';
               sreg_PLAY<=go_left;
            elsif ((led_state(1)='1') and (hit_right='1')) or   
                                             -- early hit
               ((led_state(0)='1') and ( hit_right='0' )) then   
                                             -- fail to hit on time
               score_left <= score_left + 1; 
                             -- turns all the LEDs on, for dynamic light
               led_state <= "11111111";
               sreg_PLAY<=m_lost_right;
            else                             -- the ball is on the way
               led_state<='0'&led_state(7 downto 1);
                                         -- continue shifting LEDs right
--             sreg_PLAY<=go_right;          -- Keep old state (default)
            end if;

         when go_left =>
            if (led_state(7)='1' ) and ( hit_left='1' ) then   
                                             -- correct return
               led_state<='0'&led_state(7 downto 1);
               sreg_PLAY<=go_right;
            elsif ((led_state(6)='1') and (hit_left='1')) or   
                                             -- early hit 
                  (( led_state(7)='1') and (hit_left='0')) then   
                                             -- fail to hit on time
               score_right <= score_right + 1;
               led_state <= "11111111";                  
                             -- turns all the LEDs on, for dynamic light
               sreg_PLAY<=m_lost_left;
            else                             -- the ball is on the way
               led_state <= led_state(6 downto 0)&'0';
                                          -- continue shifting LEDs left
--             sreg_PLAY<=go_left;           -- Keep old state (default)
            end if;

         when m_lost_right =>      -- Send messages: "Right mistake" 
            if LCD_mes_req = '0' and LCD_mes_ack ='0' then 
                                             -- no message pending
               LCD_mes_req <= '1';
               LcdCfgString(1 to LCD_lr_string'length) <= 
                                                          LCD_lr_string;
               LcdMessage(1 to Right_mistake'length) <= Right_mistake;
--             sreg_PLAY <= m_lost_right;    -- Keep old state (default)
            elsif LCD_mes_req = '1' and LCD_mes_ack ='1' then   
                                    -- the last message was acknowledged
               LCD_mes_req <= '0';           -- remove the request
               sreg_PLAY <= lost_right;
--          else
--             sreg_PLAY <= m_lost_right;    -- Keep old state (default)
            end if;

         when m_lost_left =>      -- Send messages: "Left mistake" 
            if LCD_mes_req = '0' and LCD_mes_ack ='0' then 
                                             -- no message pending
               LCD_mes_req <= '1';
               LcdCfgString(1 to LCD_lr_string'length) <= 
                                                          LCD_lr_string;
               LcdMessage(1 to Left_mistake'length) <= Left_mistake;
--             sreg_PLAY <= m_lost_left;     -- Keep old state (default)
            elsif LCD_mes_req = '1' and LCD_mes_ack ='1' then   
                                    -- the last message was acknowledged
               LCD_mes_req <= '0';           -- remove the request
               sreg_PLAY<=lost_left;
--          else
--             sreg_PLAY <= m_lost_left;     -- Keep old state (default)
            end if;

         when lost_right =>
            if (led_state(0)='0') then--the dynamic light sequence ended
               if score_left = 9 then        -- the game ended
                  sreg_PLAY<=m_left_won;
               else        -- the game continues, left player will serve
                  led_state <= "10000000";
                                        -- set the initial ball position
                  sreg_PLAY<=m_start_left;
               end if;
            else
               led_state <= '0' & led_state(7 downto 1);   
                                  -- continue the dynamic light sequence
--             sreg_PLAY<=lost_right;        -- Keep old state (default)
            end if;

         when lost_left =>
            if (led_state(7)='0') then  -- dynamic light sequence ended
               if (score_right = 9) then -- the geame ended
                  sreg_PLAY<=m_right_won;
               else        -- the game continues, left player will serve
                  led_state <= "00000001";
                                        -- set the initial ball position
                  sreg_PLAY<=m_start_right;
               end if;
            else
               led_state <= led_state(6 downto 0)&'0';   
                                  -- continue the dynamic light sequence
--             sreg_PLAY<=lost_left;         -- Keep old state (default)
            end if;

         when m_right_won =>      -- Send message: "Right player won"
            if LCD_mes_req = '0' and LCD_mes_ack ='0' then 
                                             -- no message pending
               LCD_mes_req <= '1';
               LcdCfgString(1 to LCD_ur_string'length) <= 
                                                          LCD_ur_string;
               LcdMessage(1 to Right_player_won'length) <= 
                                                       Right_player_won;
--             sreg_PLAY <= m_right_won;     -- Keep old state (default)
            elsif LCD_mes_req = '1' and LCD_mes_ack ='1' then   
                                    -- the last message was acknowledged
               LCD_mes_req <= '0';           -- remove the request
               sreg_PLAY <= m_game_over;
--          else
--             sreg_PLAY <= m_right_won;     -- Keep old state (default)
            end if;

         when m_left_won =>      -- Send messages: "Left player won"
            if LCD_mes_req = '0' and LCD_mes_ack ='0' then 
                                             -- no message pending
               LCD_mes_req <= '1';
               LcdCfgString(1 to LCD_ur_string'length) <= 
                                                          LCD_ur_string;
               LcdMessage(1 to Left_player_won'length) <= 
                                                        Left_player_won;
--             sreg_PLAY <= m_left_won;      -- Keep old state (default)
            elsif LCD_mes_req = '1' and LCD_mes_ack ='1' then
                                    -- the last message was acknowledged
               LCD_mes_req <= '0';           -- remove the request
               sreg_PLAY <= m_game_over;
--          else
--             sreg_PLAY <= m_left_won;      -- Keep old state (default)
            end if;
         
         when m_Game_over =>      -- Send messages: "Press B to start" 
            if LCD_mes_req = '0' and LCD_mes_ack ='0' then 
                                             -- no message pending
               LCD_mes_req <= '1';
               LcdCfgString(1 to LCD_lr_string'length) <= 
                                                          LCD_lr_string;
               LcdMessage(1 to Press_B_to_start'length) <= 
                                                       Press_B_to_start;
--             sreg_PLAY <= m_Game_over;     -- Keep old state (default)
            elsif LCD_mes_req = '1' and LCD_mes_ack ='1' then
                                    -- the last message was acknowledged
               LCD_mes_req <= '0';           -- remove the request
               sreg_PLAY <= Game_over;
--          else
--             sreg_PLAY <= m_Game_over;     -- Keep old state (default)
            end if;

         when Game_over =>
            if (start_game='1') then--the "start Game" button is pressed
               score_right <= 0;             -- initialize the scores
               score_left <= 0;
               if random_serv = '0' then   -- decide which player serves
                  led_state <= "10000000";     
                                        -- set the initial ball position
                  sreg_PLAY <= m_start_left;
               else
                  led_state <= "00000001"; 
                                        -- set the initial ball position
                  sreg_PLAY <= m_start_right;
               end if;
--          else
--             sreg_PLAY <= Game_over;       -- Keep old state (default)
            end if;

         when others =>   
            led_state <= "00000000";
            score_left <= 0;
            score_right <= 0;
            LCD_mes_req <= '0';
            sreg_PLAY <= m_Start_up;
      end case;
   end if;
end process;

end behavioral;
