------------------------------------------------------------------------
--  LcdSwCtrl.vhd -- Digilent Data Bus Controller for Digilab IO2
------------------------------------------------------------------------
--  Author : Mircea Dabacan
--           Copyright 2004 Digilent, Inc.
------------------------------------------------------------------------
-- Software version: Xilinx ISE 6.2.03i
--                   WebPack
------------------------------------------------------------------------
-- This is the source file for the LcdSwCtrl component, 
-- provided by the Digilent Component Library.

--   This file contains the design for a Data Bus Controller to be  
-- included in a user project and downloaded to a Digilent FPGA System 
-- board.
-- The module implements an interface for the Digilab Digital IO2 board.
--   It allows the user application to access the Digilab IO2 hardware 
-- resources in a very simple manner, making the DIO2 data bus (standard
-- CPLD configuration) transparent. The DIO2 resources are seen as each 
-- signal would be individually connected to an FPGA pin. There are two
-- classes of resources available on Digilent IO2 board:
   -- non-protocol devices (7-segment display, LEDs, buttons, switches).
      -- A "non-protocol" device can be written/read repetitively an 
      -- indeterminate number of times, even if the  sent/received
      -- information doesn't change. A register stores the information. 
      -- LcdSwCtrl periodically updates its' content, regardless the 
      -- fact that the input changed or not since last time. The maximum
      -- transfer latency is right the update period and represents the 
      -- maximum time from the moment when the input changes until the 
      -- output reflects the change.
   -- Lcd - microcontroller interfaced - handshake protocol required. 
      -- A "handshake protocol" requiring device (i.e. the LCD) needs a 
      -- specific information to be written/read just once (for example,  
      -- if the "A" ASCII code would be sent repetitively to the LCD, it 
      -- would be displayed multiple times on the LCD screen). Because 
      -- the controller and the user application are asynchronous to  
      -- each other, a protocol is required to ensure that any byte of 
      -- information is sent for sure once and only once. The user 
      -- application launches a transfer Request and waits for the 
      -- LcdSwCtrl to Acknowledge it (that means the request is being 
      -- processed). The user application removes now the Request. 
      -- The user application is allowed to activate a new request only 
      -- when the LcdSwCtrl releases the Acknowledge (meaning that the 
      -- requested transfer completed)
      -- A similar handshake protocol synchronizes the LcdSwCtrl to the 
      -- LCD microcontroller. After receiving a byte, the LCD controller
      -- needs a time to process it. The processing time depends on the 
      -- byte significance.
      -- Generally, configuration bytes need more time to be processed.
      -- During all the processing time, read status cycles are allowed  
      -- and the most significant bit of the read byte shows if the LCD 
      -- controller is busy.
      -- LcdSwCtrl checks the LCD status and removes the acknowledge 
      -- only when the LCD microcontroller is ready to receive new data.

------------------------------------------------------------------------
--  Port definitions
------------------------------------------------------------------------
-- LcdSwCtrl interfaces between the user application and the DIO2 Board.
-- On the user application side, the input signals (representing buttons
-- and switches) reflect almost instantaneous the state of the IO2 
-- devices. The maximal delay (15 ck periods) represent the polling 
-- cycle period, as described below. 
-- The output signals (LEDs and 7 segment display) are also transmitted  
-- to the physical devices with same maximal delay.
-- The LCD data transmission attempts have same periodicity, but the LCD
-- microcontroller itself can delay the data transfer.

-- The user application signals are:
   -- non-protocol devices (7-segment display, LEDs, buttons, switches)
      -- user application generated signals:
         --   disp7s(15 downto 0);
              -- four digit hexadecimal number for the 7 seg display
         --   LED_state(15 downto 0);
              --   16 LEDs, LD0 (left) to LD15 (right) 
      -- user application read signals:
         --   btn(14 downto 0);         -- DIO2 keypad buttons
         --   sw(7 downto 0);           -- DIO2 switches
   -- Lcd - microcontroller interfaced - handshake protocol required. 
      -- Handshake signals dealing with Lcd interface (request signals,
      -- generated by user application, acknowledge signal, 
      -- generated by LcdSwCtrl)
         -- ReqLcd    -- Lcd cycle request
         -- InitReqLcd -- Lcd Register Select request('0' - data, 
                       --                             '1' - instruction)
         -- RwReqLcd -- Lcd Read/Write request ('1' - read, 
                       --                         '0' - write)
         -- AckLcd    -- Lcd cycle acknowledge
      --  Lcd images:
         -- byteLcdWrite(7 downto 0);   -- the byte to send to Lcd
         -- byteLcdRead(7 downto 0)     -- the byte read from Lcd


-- On the DIO2 side, the module implements the signals and time 
-- sequences required to control the DIO2 data bus, as described in the 
-- Digilab Digital IO2 Reference Manual (standard CPLD configuration).
-- It generates a polling procedure to access all the hardware resources
-- provided by DIO2.

-- The DIO2 bus consists in:
   -- an 8 bit bi-directional data bus:
      -- dbLcd(7 downto 0)
   -- a 6 bit address bus (unidirectional, generated by system board):
      --   IO2_adr(5 downto 0)
   -- a 3 bit control bus (unidirectional, generated by system board):
      -- IO2_CS - CPLD Chip Select
      -- IO2_WE - CPLD Write Enable
      -- IO2_OE - CPLD Output Enable.

-- According to the Digilent Digital IO2 board Reference Manual, 
-- the hardware resources are accessible through the system bus
-- as byte-wide read or write registers:
   -- Address   Resource               Direction
   -- 0         Buttons  7 - 0         read
   -- 1         Buttons 14 - 8         read
   -- 2         Switches 8 - 1         read
   -- 3         not used
   -- 4         LEDs 7 - 0             write
   -- 5         LEDs F - 8             write
   -- 6         7s digits 1 & 2        write
   -- 7         7s digits 3 & 4        write

-- The dbLcd is also used by the IO2 Lcd, together with specific 
-- control signals:
   --   EnLcd  - Lcd enable (active high)
   --   RsLcd - Lcd Register Select ('1' - data, '0' - instruction)
   --   RwLcd - Lcd Read/Write ('1' - read, '0' - write)
   
------------------------------------------------------------------------
--  Internal signals
------------------------------------------------------------------------
   -- dbLcd_o (7 downto 0) - the value prepared to output on the dbLcd
   -- OeDbLcd - the data bus Output Enable for the FPGA chip. 
   -- stDbCtrlCur(4 downto 0)  - the FSM state register
   -- stDbCtrlNext(4 downto 0) - the next state

------------------------------------------------------------------------
--  Behavioral description
------------------------------------------------------------------------
-- The LcdSwCtrl module browses all the system bus used addresses, 
-- generating appropriate (read or writes) cycles. 
-- Read data is stored in internal registers.
-- Data to be written (stored in user application registers) is sent 
-- over the bus.
-- A complete sequence includes, in this order:
   -- write address 4 (LEDs 7 - 0)            
   -- write address 5 (LEDs F - 8)            
   -- write address 7 (7s digits 3 & 4)            
   -- write address 6 (7s digits 1 & 2)            
   -- read  address 2 (Switches 8 - 1)
   -- read  address 0 (Buttons  7 - 0)
   -- read  address 1 (Buttons  14 - 8)
   -- LCD cycle 

-- Each state takes a ck period, a write cycle takes two ck periods 
-- (2 states) and a read cycle takes 1 ck period (1 state). 
-- An LCD cycle takes four ck periods (4 states).
-- The whole dbLcd cycle takes 15 ck periods.
-- Using a 3.125MHz clock, the dbLcd cycle frequency is 208.33KHz.
-- The LcdSwCtrl latency is the cycle period = 4.8us. 

-- IMPORTANT NOTE: To keep the compatibility between multiple similar 
-- dbIOxCtrl variants, the symbols CS, OE, WE are encoded in the state
-- register and in the time diagrams below as being active HIGH
-- (positive logic). For the Digilab Digital IO2 board this is the case-
-- active high signals: IO2_CS, IO2_OE, IO2_WE. They are assigned at the
-- end of the file as:
-- IO2_CS <= stDbCtrlCur(0); -- IO2_CS active (HIGH) except for first 
                             -- and last states in a LCD cycle 
-- IO2_OE <= stDbCtrlCur(1); -- IO2_OE active (HIGH) during all read 
                             -- states
-- IO2_WE <= stDbCtrlCur(2); -- IO2_WE active (HIGH) during all write 
                             -- states

-- Each cycle type is described below: 

  -- A write cycle is generated for each IO2_adr in range 4 to 7. 
  -- The actual order is (IO2_adr in brackets):
  -- LedLo (4), LedHi (5), SsegHi (7), SsegLo (6).
  -- Each write cycle consists in two subsequent states, a "write" one
  -- and an "idle" one.
  -- The state machine activates the dbLcd and generates the control
  -- signals:
    -- CS is active during the entire cycle,
    -- OE is inactive during the entire cycle, 
    -- WE is generated active for Wr states and inactive for Idle states
    -- dbLcd value is generated by the FPGA chip, depending on IO2_adr,
       -- as shown above,
    -- WE signal forces the CPLD to latch the value of dbLcd at the end
       -- of stWrXx state.

-- stDbCtrlCur ...| stWrXx | stIdXx |...
--             _ _ _________________ _ _
-- CS          _ _|                 |_ _ active (H) during write cycle
--             _ _                   _ _
-- OE          _ _|_________________|_ _ inactive (L) during write cycle
--             _ _ ________          _ _
-- WE          _ _|        |________|_ _ active (H) during Wr state 
--             _ _ _________________ _ _
-- OeDbLcd     _ _|                 |_ _ internal signal 
--             _ _ _________________ _ _
-- dbLcd       _ _|_________________|_ _ write data (FPGA generated)

-- Notes: WE falling edge stores data in CPLD registers
--        OeDbLcd activates the dbLcd, when HIGH


  -- A READ CYCLE is generated for each IO2_adr in range 0 to 2.
  -- The actual order is (IO2_adr in brackets):
  -- Swt (2), BtnLo (0), BtnHi (1).
  -- Each write cycle consists in a single "read" state.
  -- The state machine generates the control signals and reads dbLcd:
    -- CS is active during the entire cycle,
    -- WE is inactive during the entire cycle, 
    -- OE is generated active for Rd states,
    -- dbLcd value is sampled at the end of the read state.

-- stDbCtrlCur ...| stRdXx  |...
--             _ _ _________ _ _
-- CS          _ _|         |_ _ active (HIGH) during read cycle
--             _ _ _________ _ _
-- OE          _ _|         |_ _active (HIGH) during read cycle
--             _ _           _ _
-- WE          _ _|_________|_ _inactive (LOW) during Rd cycle
--             _ _           _ _
-- OeDbLcd     _ _|_________|_ _internal signal
--             _ _ _________ _ _
-- dbLcd       _ _|_________|_ _write data (generated by IO2 Board)

-- Note: OeDbLcd activates the dbLcd, when HIGH

-- Lcd cycle (and protocol)

  -- An LCD read, write or passive cycle is generated in a bus sequence.
  -- The type of the Lcd cycle is decided based on the request signals,
  -- generated by the user application, as well as acknowledge signal, 
  -- generated by the LcdSwCtrl:

-- ReqLcd  AckLcd    RwReqLcd InitReqLcd Cycle type        AckLcd 
--                                                         next value
--    0       0          X       X       passive cycle     0 (unchanged)
--       (no Lcd cycle request)
--    1       0          0       1       write Lcd command     1   
--    1       0          0       0       write Lcd data        1
--    1       0          1       1       read Lcd status       1
--    1       0          1       0       read Lcd data         1
--    1       1          X       X       passive cycle     1 (unchanged)
--       (the requested Lcd cycle already complete)
--    0       1          X       X       passive cycle     0 if LCD Busy
--       (the current request ended)                         flag is 0
--       (acknowledge is released if the LCD controller isn't more busy)

  -- As the table above shows, ReqLcd and AckLcd signals implement a 
  -- handshake protocol, ensuring that a single Lcd cycle is generate as
  -- result of a single request.

--             ___________________                         generated 
-- ReqLcd  ___|                   |_______________________ by user 
--                                                         application
--                 _______________________________             
-- AckLcd  _______|                               |_______ generated 
--          0 | 1 | 2 | 3 |...| 3 | 4 |...| 4 | 5 | 0 |...|by LcdSwCtrl

-- The time diagram above explains the handshake protocol:
  -- 0 - no Lcd cycle request pending: the user application doesn't 
  --     need Lcd access; LcdSwCtrl generates passive LCD cycles.
  -- 1 - LcdSwCtrl latency: some other read or write cycles running.
  -- 2 - LcdSwCtrl generates the requested Lcd cycle; AckLcd goes high.
  -- 3 - the user application latency: ReqLcd goes low;
  --     LcdSwCtrl generates passive LCD cycles.
  -- 4 - LcdSwCtrl passive cycles check the LCD Busy flag, 
  --     which is found active (HIGH)
  -- 5 - LcdSwCtrl passive cycle check the LCD Busy flag, 
  --     which is found inactive (LOW); As result, AckLcd goes low.

   -- The time diagrams below explains the actual Lcd write, read (2 in
   -- the time diagram above)or passive cycle. 
   -- During an Lcd cycle, the IO2 control signals (CS, OE, WE) are 
   -- permanently inactive:
     -- The state machine activates the dbLcd and generates the control
     -- signals:
       -- RsLcd is the image of InitReqLcd: Lcd Register Select 
          -- ('0' - data, '1' - instruction)
       -- RwLcd is the image of RwReqLcd: Lcd Read/Write 
          -- ('1' - read, '0' - write)
       -- EnLcd - Lcd enable (active high pulse) 

      -- An LCD WRITE CYCLE consists in four subsequent states:

-- stDbCtrlCur ...|stWrLcdX0|stWrLcdX1|stWrLcdX2|stWrLcdX3|...
--             _ _ _______________________________________ _ _
-- RsLcd       _ _|_______________________________________|_ _  Note 1.
--             _ _                                         _ _
-- RwLcd       _ _|_______________________________________|_ _  Note 2.
--             _ _           ___________________           _ _
-- EnLcd       _ _|_________|                   |_________|_ _  Note 3.
--             _ _ _______________________________________ _ _
-- OeDbLcd     _ _|                                       |_ _  Note 4.
--             _ _ _______________________________________ _ _
-- dbLcd       _ _|_______________________________________|_ _  Note 5.
--             _ _ _ _ _ _ _ ___________________ _ _ _ _ _ _ _
-- LCD bus     _ _ _ _ _ _ _|___________________|_ _ _ _ _ _ _  Note 6. 
--                                                                  

-- Note 1. '1' - data, '0' - instruction
-- Note 2. '0'- write
-- Note 3. Active (HIGH) for the middle states
-- Note 4. Internal signal, activates the dbLcd, when HIGH
-- Note 5. Write data (FPGA generated)
-- Note 6. Between CPDL and LCD microcontroller (CPLD generated)
-- (enabled by CS and WE, (latched by the LCD on the EnLcd falling edge)

  -- An LCD READ CYCLE consists in two subsequent states, a "read" one 
  -- and an "idle" one.

-- stDbCtrlCur   ...|stRdLcdX0|stRdLcdX1|stRdLcdX2|stRdLcdX3|...
--             _ _ _______________________________________ _ _
-- RsLcd       _ _|_______________________________________|_ _ Note 1.
--             _ _ _______________________________________ _ _
-- RwLcd       _ _|                                       |_ _ Note 2.
--             _ _           ___________________           _ _
-- EnLcd       _ _|_________|                   |_________|_ _ Note 3.
--             _ _ _______________________________________ _ _
-- OeDbLcd     _ _|                                       |_ _ Note 4.
--             _ _ _ _ _ _ _ ___________________ _ _ _ _ _ _ _
-- IO2 bus     _ _ _ _ _ _ _|___________________|_ _ _ _ _ _ _ Note 5.
--             _ _ _ _ _ _ _ ___________________ _ _ _ _ _ _ _
-- LCD_db      _ _ _ _ _ _ _|___________________|_ _ _ _ _ _ _ Note 6.
--                                                                     
-- Note 1. '1' - data, '0' - instruction
-- Note 2. '1'- read
-- Note 3. Active (HIGH) for the middle states
-- Note 4. Internal signal, activates the dbLcd, when HIGH
-- Note 5. Between CPDL and LCD microcontroller (LCD generated)
--         (enabled by RwLcd and EnLcd)
-- Note 6. Read data (CPLD generated)
--         (latched by the LcdSwCtrl at the end of stRdLcdX2)


  -- An Lcd passive cycle is similar to An Lcd status read cycle, 
  -- except that the read data is not latched to byteLcdRead.
  -- Instead, the most significant bit of the read data is checked 
  -- (if ReqLcd is inactive also): 
    -- If '1' (the LCD controller is busy), AckLcd is maintained active 
      -- (high) and consequently the next LCD cycle will be also a 
      -- passive one.
    -- If '0' (the LCD controller is ready), AckLcd is inactivated 
      -- (low). Consequently, the use application is allowed to request
      -- a new byte transfer. If this does not happen, the next Lcd 
      -- cycle(s) will also be passive. When the ReqLcd activates, the
      -- proximal LCD cycle will be a write or read one (upon request
      -- type). 

------------------------------------------------------------------------
-- Revision History:
--   12/21/2004(MirceaD): created
--   03/24/2005(MirceaD): formatted
------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;     -- defines std_logic types
use IEEE.std_logic_arith.all;    -- def. conv_std_logic_vector function.
use IEEE.std_logic_unsigned.all; -- def. add std_logic_vector operation.


entity LcdSwCtrl is
    port (
      ck : in STD_LOGIC;            -- db state machine clock (3.125MHz)
      reset : in std_logic;

--   Other Phoenix resources:
      sw  : out std_logic_vector (7 downto 0); -- sw state
 
      RamData: in  std_logic_vector(63 downto 0);
      RamAdr:  out std_logic_vector(6 downto 0);


--   Lcd signals:
      dbLcd  : inout std_logic_vector (7 downto 0); -- data bus
      EnLcd : out std_logic:='0';        -- Lcd enable (active high)
      RsLcd : out std_logic;       -- Lcd Register Select 
                                      -- ('1' - data, '0' - instruction)
      RwLcd : out std_logic;       -- Lcd Read/Write
                                      -- ('1' - read, '0' - write)
      Cs1Lcd : out std_logic;       -- Lcd Right Chip Select (active high)
      Cs2Lcd : out std_logic;       -- Lcd Left Chip Select (active high)
      RetLcd : out std_logic       -- Lcd Reset (active low)

--  Lcd images:
--      byteLcdWrite : in std_logic_vector (7 downto 0);-- byte for Lcd
--      byteLcdRead : out std_logic_vector (7 downto 0) -- byte from Lcd
      );
end LcdSwCtrl;

architecture behavioral of LcdSwCtrl is

------------------------------------------------------------------------
-- Constant and Signal Declarations
------------------------------------------------------------------------
-- The following constants define the state codes for the IO5 Data Bus 
-- control state machine. This state machine controls the sequence of 
-- operations needed to perform data bus cycles for all the IO resources
-- of the Digilab IO5 board. 
-- States are such a way assigned to directly encode the output signals:
-- IO2_adr(3 downto 0), IO2_WE, IO2_OE, IO2_E - active HIGH.
-- That eliminates output logic and generates glitch free outputs.
-- The assignment also tries to minimize the number of state register  
-- bits changing at state transitions (Grey code - like)
--constant   stWrLedLo:      std_logic_vector(5 downto 0) := "100100"; 
--constant   stIdLedLo:      std_logic_vector(5 downto 0) := "100000"; 
--constant   stWrLedHi:      std_logic_vector(5 downto 0) := "101100"; 
--constant   stIdLedHi:      std_logic_vector(5 downto 0) := "101000"; 
--constant   stWrSsegHi:     std_logic_vector(5 downto 0) := "111100"; 
--constant   stIdSsegHi:     std_logic_vector(5 downto 0) := "111000"; 
--constant   stWrSsegLo:     std_logic_vector(5 downto 0) := "110100"; 
--constant   stIdSsegLo:     std_logic_vector(5 downto 0) := "110000"; 
constant   stRdSwt:        std_logic_vector(5 downto 0) := "010010"; 
--constant   stRdBtnLo:      std_logic_vector(5 downto 0) := "000010"; 
--constant   stRdBtnHi:      std_logic_vector(5 downto 0) := "001010"; 
constant   stWrLcdCmd0:    std_logic_vector(5 downto 0) := "000100"; 
constant   stWrLcdCmd1:    std_logic_vector(5 downto 0) := "000101"; 
constant   stWrLcdCmd2:    std_logic_vector(5 downto 0) := "010101"; 
constant   stWrLcdCmd3:    std_logic_vector(5 downto 0) := "010100"; 
constant   stRdLcdStatus0: std_logic_vector(5 downto 0) := "100010"; 
constant   stRdLcdStatus1: std_logic_vector(5 downto 0) := "100011"; 
constant   stRdLcdStatus2: std_logic_vector(5 downto 0) := "110011"; 
constant   stRdLcdStatus3: std_logic_vector(5 downto 0) := "110010"; 
constant   stWrLcdData0:   std_logic_vector(5 downto 0) := "001100"; 
constant   stWrLcdData1:   std_logic_vector(5 downto 0) := "001101"; 
constant   stWrLcdData2:   std_logic_vector(5 downto 0) := "011101"; 
constant   stWrLcdData3:   std_logic_vector(5 downto 0) := "011100"; 
constant   stRdLcdData0:   std_logic_vector(5 downto 0) := "101010"; 
constant   stRdLcdData1:   std_logic_vector(5 downto 0) := "101011"; 
constant   stRdLcdData2:   std_logic_vector(5 downto 0) := "111011"; 
constant   stRdLcdData3:   std_logic_vector(5 downto 0) := "111010"; 

-- State register and next state for the FSM
signal stDbCtrlCur : std_logic_vector(5 downto 0) := stRdSwt;
signal stDbCtrlNext : std_logic_vector(5 downto 0):= stRdSwt;  

-- The attribute lines below prevent the ISE compiler to extract and 
-- optimize the state machines.
-- WebPack 5.1 doesn't need them (the default value is NO)
-- WebPack 6.2 has the default value YES, so without these lines would 
-- "optimize" the state machines.
-- Although the overall circuit would be optimized, the particular goal 
-- of "glitch free output signals" may not be reached. 
-- That is the reason of implementing the state machine as described in 
-- the constant declarations above. 

attribute fsm_extract : string;
attribute fsm_extract of stDbCtrlCur: signal is "no"; 
attribute fsm_extract of stDbCtrlNext: signal is "no"; 

attribute fsm_encoding : string;
attribute fsm_encoding of stDbCtrlCur: signal is "user"; 
attribute fsm_encoding of stDbCtrlNext: signal is "user"; 

attribute signal_encoding : string;
attribute signal_encoding of stDbCtrlCur: signal is "user"; 
attribute signal_encoding of stDbCtrlNext: signal is "user";

-- Data bus control signals   
   signal OeDbLcd: std_logic;       -- data bus output enable
   signal dbLcd_o: std_logic_vector(7 downto 0);   
                                    -- output value for data bus

-- Handshake signals dealing with Lcd interface
--signal ReqLcd:   std_logic;       -- Lcd cycle request
signal InitReqLcd: std_logic := '1';    -- Lcd Register Select request
                                      -- ('1' - data, '0' - instruction)
signal ColAdrReqLcd: std_logic := '1';    -- Lcd columna address required
                                      -- ('1' - required, '0' - not)
--      RwReqLcd : in std_logic;    -- Lcd Read/Write request
                                      -- ('1' - read, '0' - write)
signal AckLcd: std_logic := '0';    -- Lcd cycle acknowledge

-- Signals to control the cfg/data sequence
signal cntByte: integer range 0 to 127 := 0; 
       -- Counter for cfg or data bytes
signal cntRow: integer range 0 to 7 := 0; -- Counter for LCD row
signal cntPage: std_logic:='0'; -- page counter
-- LCD configuration string
--type config_string_type is array (0 to 5)  --                              Debug???
type config_string_type is array (0 to 4)  --                              Debug???
                                 of std_logic_vector(7 downto 0);
constant LcdInitString : config_string_type:=
--        (x"40", -- Set Y (column) address to 0
--         x"B8", -- Set X (row) address to 0
--         x"C0", -- Set display line to 0
--         x"3f"); -- Set Display ON


--initial
        (x"3f", -- Set Display ON   
         x"3f", -- Set Display ON   
         x"c0", -- Set display line to 0
         x"40", -- Set Y (column) address to 0
         x"B8"); -- Set X (row) address to 0
--         x"C0", -- Set display line to 0
--         x"3f"); -- Set Display ON

-- try

--(x"e2",x"a1",x"c0",x"a3",x"2f",x"26",x"81",x"1c",x"40",x"af");
signal LcdCfgString : config_string_type:= LcdInitString;
signal LcdCfgByte: std_logic_vector(7 downto 0);


signal GATA: std_logic:='0'; -- DEBUG???



--Actual architecture begin
begin
------------------------------------------------------------------------
-- LCD/sw Bus Controller State Machine
------------------------------------------------------------------------

State_register:   process (ck, reset)
   begin
      if reset = '1' then
         stDbCtrlCur <= stRdSwt; 
      elsif ck'Event and ck = '1' then
         stDbCtrlCur <= stDbCtrlNext;
      end if;
   end process;

next_state_logic:   process (stDbCtrlCur)
   begin
      case stDbCtrlCur is
      -- Last Read cycle. Decides what type of LCD cycle follows
         when stRdSwt =>
--            if ReqLcd = '1' and           -- Active Request


IF GATA = '1' then                           --DEBUG???
      stDbCtrlNext <= stRdSwt; -- NO MORE LCD CYLES EVER
ELSE

            if AckLcd = '0' then -- The LCD was found not busy
               if InitReqLcd = '1' or ColAdrReqLcd = '1' then  -- Data register requested
--                  if RwReqLcd = '0' then -- Write cycle requested
                     stDbCtrlNext <= stWrLcdCmd0;
--                  else
--                     stDbCtrlNext <= stRdLcdData0;-- Read LCD Data cycle
--                     null;  -- LCD data read not suported
--                  end if;
               else                 -- Command/Status register requested
--                  if RwReqLcd = '0' then   -- Write cycle requested
                     stDbCtrlNext <= stWrLcdData0;--Write LCD Data cycle
                               -- Write LCD Command cycle
--                  else
--                     stDbCtrlNext <= stRdLcdStatus0;
                               -- Read LCD Status cycle (user requested)
--                     null;-- user requested LCD status read not suported
--                  end if;
               end if;
            else
                  stDbCtrlNext <= stRdLcdStatus0;
                   -- Read LCD Status cycle (to check the LCD busy flag)
            end if;


END IF;




   -- Write LCD Command cycle
         when stWrLcdCmd0 =>
            stDbCtrlNext <= stWrLcdCmd1;
         when stWrLcdCmd1 =>
            stDbCtrlNext <= stWrLcdCmd2;
         when stWrLcdCmd2 =>
            stDbCtrlNext <= stWrLcdCmd3;
         when stWrLcdCmd3 =>
            stDbCtrlNext <= stRdSwt;
   -- Read LCD Status cycle
         when stRdLcdStatus0 =>
            stDbCtrlNext <= stRdLcdStatus1;
         when stRdLcdStatus1 =>
            stDbCtrlNext <= stRdLcdStatus2;
         when stRdLcdStatus2 =>
            stDbCtrlNext <= stRdLcdStatus3;
         when stRdLcdStatus3 =>
            stDbCtrlNext <= stRdSwt;
   -- Write LCD Data cycle
         when stWrLcdData0 =>
            stDbCtrlNext <= stWrLcdData1;
         when stWrLcdData1 =>
            stDbCtrlNext <= stWrLcdData2;
         when stWrLcdData2 =>
            stDbCtrlNext <= stWrLcdData3;
         when stWrLcdData3 =>
            stDbCtrlNext <= stRdSwt;
   -- Read LCD Data cycle
--         when stRdLcdData0 =>
--            stDbCtrlNext <= stRdLcdData1;
--         when stRdLcdData1 =>
--            stDbCtrlNext <= stRdLcdData2;
--         when stRdLcdData2 =>
--            stDbCtrlNext <= stRdLcdData3;
--         when stRdLcdData3 =>
--            stDbCtrlNext <= stRdSwt;
         when others =>                     -- unknown states
            stDbCtrlNext <= stRdSwt;
      end case;
   end process;

------------------------------------------------------------------------
-- Assigning read registers:
------------------------------------------------------------------------
read_reg: process (ck)      -- synchronous process
   begin

      if (ck'event and ck='1') then            -- ck rising edge
         if stDbCtrlCur = stRdSwt then
            sw<=dbLcd;       -- switches 7..0 read at IO2_adr 2
--         elsif stDbCtrlCur = stRdBtnLo then
--            btn(7 downto 0) <= dbLcd;
                              -- buttons 7..0 read at IO2_adr 0
--         elsif stDbCtrlCur = stRdBtnHi then
--            btn(14 downto 8) <= dbLcd(6 downto 0);
                              -- buttons 14..8 read at IO2_adr 1
--         elsif (stDbCtrlCur = stRdLcdStatus2 and 
--                ReqLcd = '1' and AckLcd = '0') or 
--                AckLcd = '0') or 
--               stDbCtrlCur = stRdLcdData2 then
--            byteLcdRead<=dbLcd;  -- Lcd status or data byte read
         end if;
      end if;      
   end process;


------------------------------------------------------------------------
-- command/data sequence:
------------------------------------------------------------------------
dataSequence: process(ck, stDbCtrlCur)
  begin
  if ck'event and ck = '1' then
    if stDbCtrlCur = stWrLcdCmd3 then -- a command byte has been written
--       if cntByte /= 3 then -- there are some more command bytes to send



     if InitReqLcd = '1' then-- init in process
     


       if cntByte /= 4 then -- there are some more command bytes to send      DEBUG???
         cntByte <= cntByte + 1; -- next command byte
         InitReqLcd <= '1';      -- CONTINUE CMD bytes REDUNDANT
       else -- no more command bytes
         cntByte <= 0; -- prepared for data bytes
         InitReqLcd <= '0';      -- prepared for data bytes
       end if;



     elsif  ColAdrReqLcd = '1' then  -- column address required
         ColAdrReqLcd <= '0' ; --address sent, data follows
     end if;



    elsif stDbCtrlCur = stWrLcdData3 then --a data byte has been written



--       if cntByte /= 127 then -- there are some more data bytes to send
       if cntByte /= 63 then -- there are some more data bytes to send     64 for a chip, 128 for both




         cntByte <= cntByte + 1; -- next data byte
         InitReqLcd <= '0';      -- CONTINUE DATA bytes REDUNDANT


         ColAdrReqLcd <= '1' ; -- data sent, address follows



       else -- no more data bytes

--GATA <= '1'; -- DEBUG???

         cntByte  <= 0; -- prepared for command bytes

   if cntRow/=7 then
         cntRow <= cntRow + 1; -- next row
         InitReqLcd <= '1';      -- prepared for command bytes
   else
         InitReqLcd <= '1'; -- init sequence requested
         cntPage <= not cntPage; -- the other half of LCD
         cntRow <= 0;       -- first row
--GATA <= '1'; -- DEBUG???
   end if;


       end if;
    end if;
  end if;
  end process;
------------------------------------------------------------------------
-- Handshake protocol signal (AckLcd):
------------------------------------------------------------------------
handshake: process (ck, reset)      -- synchronous process
   begin
      if reset = '1' then
         AckLcd <= '0';                  -- AckLcd goes inactive
      elsif (ck'event and ck='1') then        -- ck rising edge
         if (stDbCtrlCur = stWrLcdCmd0) then -- and 
--             ReqLcd = '1' and AckLcd = '0') then-- requested LCD cycle
--             AckLcd = '0') then          -- LCD was found not busy
            AckLcd <= '1';               -- AckLcd goes active



         ELSif (stDbCtrlCur = stWrLcdData0) then -- and 
            AckLcd <= '1';               -- AckLcd goes active



         elsif stDbCtrlCur = stRdLcdStatus2 and
               dbLcd(7) = '0' then       -- LCD controller is ready
--               dbLcd(7) = '0' and            -- LCD controller is ready
--               ReqLcd = '0' then             -- Lcd request ended 
            AckLcd <= '0';               -- AckLcd goes inactive
         end if;
      end if;      
   end process;

------------------------------------------------------------------------
-- Controlling the three state of dbLcd (write, read or idle)
------------------------------------------------------------------------
   OeDbLcd <= stDbCtrlCur(2);   
             -- OeDbLcd active (HIGH) during all write states
             -- (reduces VDD glitches)
   dbLcd <= dbLcd_o when OeDbLcd = '1' else "ZZZZZZZZ";      

   EnLcd <= stDbCtrlCur(0);
             -- EnLcd active (HIGH) during LCD cycle middle states
------------------------------------------------------------------------
-- RwLcd and RsLcd assignment:
------------------------------------------------------------------------
   RwLcd <= not stDbCtrlCur(2);
             -- IO2_WE active (LOW) during all write states
--   RsLcd <= not stDbCtrlCur(3);      -- RsLcd <= IO2_adr(0);               
   RsLcd <= stDbCtrlCur(3);      -- RsLcd <= IO2_adr(0);    PE DOS???           



--   Cs1Lcd <= '1' when (GATA = '0' and stDbCtrlCur /= stRdSwt) else '0'; -- active
----   Cs1Lcd <= '1' when (stDbCtrlCur /= stRdSwt) else '0'; -- active
--   Cs2Lcd <= '0'; -- inactive

--   Cs1Lcd <= '0';-- when (GATA = '0' and stDbCtrlCur /= stRdSwt) else '1'; -- active
--   Cs1Lcd <= '1' when (stDbCtrlCur /= stRdSwt) else '0'; -- active
--   Cs2Lcd <= '1'; -- inactive

Cs1Lcd <= not cntPage;
Cs2Lcd <= cntPage;


     RetLcd <= '0' when reset = '1' else '1'; -- active when reset         

--   RetLcd <= '1'; -- inactive          

------------------------------------------------------------------------
-- Assigning internal signals values to dbLcd (write cycles):
------------------------------------------------------------------------
--      LcdCfgString(0 to 2) <= LcdInitString(0 to 2);
      LcdCfgString(0 to 3) <= LcdInitString(0 to 3);
--      LcdCfgString(3) <= LcdInitString(3) + conv_std_logic_vector(cntByte,8);
      LcdCfgString(4) <= LcdInitString(4) + conv_std_logic_vector(cntRow,8);



   with stDbCtrlCur select 
   dbLcd_o <= 
      RamData(8*cntRow + 7 downto 8*cntRow) 
--      x"3c" 
--conv_std_logic_vector(cntByte,8)
                                      when stWrLcdData0|stWrLcdData1|
                                           stWrLcdData2|stWrLcdData3,
--      LcdInitString(conv_integer(cntByte)) when stWrLcdCmd0|stWrLcdCmd1|
--      LcdCfgString(conv_integer(cntByte)) when stWrLcdCmd0|stWrLcdCmd1|
      LcdCfgByte when stWrLcdCmd0|stWrLcdCmd1|
                                           stWrLcdCmd2|stWrLcdCmd3,
      "--------" when others;

   RamAdr <= conv_std_logic_vector(cntByte,7) when cntPage = '0' else
             conv_std_logic_vector(cntByte+64,7);


LcdCfgByte <= LcdCfgString(conv_integer(cntByte)) when InitReqLcd = '1' else 
              x"40" or conv_std_logic_vector(cntByte,8);
               
end behavioral;
