------------------------------------------------------------------------
--  Lcd_mess_Ctrl.vhd -- Digilent LCD Message Controller 
--                       for Digilab IO2 and IO5
------------------------------------------------------------------------
--  Author : Mircea Dabacan
--           Copyright 2004 Digilent, Inc.
------------------------------------------------------------------------
-- Software version: Xilinx ISE 6.2.03i
--                   WebPack
------------------------------------------------------------------------
-- This is the source file for the Lcd_mess_Ctrl component, 
-- provided by the Digilent Component Library.

--   This file contains the design for a LCD Message Controller to be  
-- included in a user project and downloaded to a Digilent FPGA System 
-- board.
-- The module implements an interface between:
   --   the user application, on one side and
   -- the dbIOxCtrl (dbIO2Ctrl or dbIO5Ctrl) component on the other side
-- The module acts as "host" for the user application "client" and 
-- as "client" for the dbIOxCtrl "host".
--   It allows the user application to send Configuration (instruction)  
-- and Data (text) messages to the Digilab IO2 or IO5 LCD in a very 
-- simple manner.
-- The DIO2 or DIO5 LCD is seen through the dbIO2Ctrl or dbIO5Ctrl 
-- component.
-- For synchronizing purposes, handshake protocols are provided at both 
-- sides:
   -- A "handshake protocol" ensures that an "information unit" is 
   -- written/read just once. At the dbIoxCtrl side, the "information
   -- unit" is a byte (a configuration (instruction) byte or an ASCII 
   -- code. At the user application side, the "information unit" is a 
   -- message, composed of a configuration string and a data (text) 
   -- string. The "client" launches a transfer Request and waits for the
   -- "host" to Acknowledge it (meaning the request is being processed).
   -- The "client" removes now the Request. The "client" is allowed to 
   -- activate a new Request only after the "host" releases the 
   -- Acknowledge (meaning that the requested transfer completed).

------------------------------------------------------------------------
--  Port definitions
------------------------------------------------------------------------
-- General purpose signals
--   ck : in std_logic;                        -- 3.125MHz clock
--   reset: in std_logic;                      -- btn1 on the main board
-- User application interface signals
--   Lcd_config_string: in config_string_type;
               -- the configuration string to be sent to the LCD
--   LCD_message:in string;    -- the text string to be sent to the LCD
--   LCD_mes_req: in std_logic; -- user appl. request to send a message
--   LCD_mes_ack: out std_logic; 
               -- acknowledge signal generated by the Lcd_Mess_Ctrl
-- dbIOxCtrl component interface signals 
--   lcd_RS_req: out std_logic; -- LCD Register Select request('1'-data,
                                --                    '0' - instruction)
--   lcd_RW_req: out std_logic; -- LCD Read/Write request ('1' - read, 
                                --                         '0' - write)
--   LCD_write_byte: out std_logic_vector (7 downto 0);
                                -- byte to send to LCD
--   LCD_req: out std_logic;    -- LCD cycle request
--   LCD_ack: in std_logic      -- LCD cycle acknowledge

-- The lcd_config_string is an array of std_logic_vector(7 downto 0);
-- This type is not recognized by default, so it needs to be declared.
-- To be used in the entity, the type must be declared somewhere before.
-- The single way to do that is using a package, which is lcd_pkg.vhd 

------------------------------------------------------------------------
--  Behavioral description
------------------------------------------------------------------------
-- signal LCD_index : integer range 1 to LCD_index_max; 
           -- the index in the configuration string or text string.

------------------------------------------------------------------------
--  Function
------------------------------------------------------------------------
-- function conv_ASCII(char:character) return STD_LOGIC_VECTOR is
-- Returns the ASCII code of the argument
   -- The "|" character is omitted in the list. As result, the 
   -- conv_ASCII function returns the code x"00" for that character.
   -- It is used as "end" mark for both configuration and text strings.

------------------------------------------------------------------------
--  Behavioral description
------------------------------------------------------------------------
-- The time diagram below explains the handshake protocol between the 
-- user application and the Lcd_Mess_Ctrl component.

--                    ______ 
-- Lcd_mes_Req     __|      |____________ generated by user application
--                        ___________
-- Lcd_mes_ack     ______|           |___ generated by Lcd_Mess_Ctrl
--                 0 | 1 | 2 | 2 |...| 0

   -- 0 - no Lcd message request pending: the user application doesn't 
      -- need Lcd access.
      -- The user application is allowed to activate Lcd_mes_Req only 
      -- when Lcd_mes_ack is inactive.
      -- Before activating Lcd_mes_Req, the user application must set 
      -- both Lcd_config_string and LCD_message. They must stay stable 
      -- while Lcd_mes_ack  is active.
   -- 1 - Lcd_Mess_Ctrl latency (max 1 ck period = 20ns).
   -- 2 - Lcd_Mess_Ctrl sends the message byte by byte to the dbIOxCtrl.
      -- (See next time diagram.)
      -- The user application can remove the Lcd_Mess_Req immediately 
      -- after Lcd_mes_ack goes active.
      -- Lcd_mes_ack remains active until two conditions are met:
         -- Lcd_mes_Req is inactive and
         -- all the message was sent to the dbIOxCtrl.

   -- The Lcd_Mess_Ctrl state machine:
      -- keeps state stMessIdle during sequence 0, 
      -- switch to state stCmdBytePrep after sequence 1, 
      -- loops states stCmdBytePrep and stCmdByteReq during each 
         -- sequence 2 until all the configuration string is sent,
      -- loops states stDataBytePrep and stDataByteReq during each 
         -- sequence 2 until all the text string is sent,
      -- keeps state stDataBytePrep until Lcd_mes_Req inactivates (if 
         -- not already inactive), 
      -- returns to state stMessIdle. 

-- The time diagram below explains the handshake protocol between the 
-- Lcd_Mess_Ctrl component and the dbIOxCtrl component (sequence 2 in 
-- the previous time  diagram). Lcd_Req is generated by Lcd_Mess_Ctrl,
-- while Lcd_ack is generated by dbIOxCtrl.

--             ___________________ 
-- Lcd_Req ___|                   |___________________________ 
--                 _______________________________
-- Lcd_ack _______|                               |___________ 
--          0 | 1 | 2 | 3 |...| 3 | 4 |...| 4 | 5 | 0 |...| 0 |

   -- 0 - no Lcd cycle request pending: Lcd_Mes_Ctrl doesn't need Lcd 
   --         access. dbIOxCtrl generates passive LCD cycles.
   -- 1 - dbIOxCtrl latency: some other read or write cycles are running
   -- 2 - dbIOxCtrl generates the requested Lcd cycle. Lcd_ack goes high
   -- 3 - Lcd_Mes_Ctrl latency: Lcd_Req goes low. dbIOxCtrl generates 
   --     passive LCD cycles.
   -- 4 - dbIOxCtrl passive cycles. dbIOxCtrl waits for two conditions 
   --     to be met simultaneously: 
      -- Lcd_Req = inactive
      -- LCD busy status bit = inactive (meaning that the LCD completed
         -- the internal operations and is ready for a new byte). 
   -- 5 - dbIOxCtrl passive cycle. dbIOxCtrl found true the two 
   --     conditions above. As result, Lcd_ack goes low.

   -- While sending the configuration message, the Lcd_Mess_Ctrl state
   -- machine:
      -- keeps state stCmdBytePrep during sequences 4 and 5, 
      -- switch to state stCmdByteReq after sequence 0 and keeps the 
         -- state during sequence 1, 
      -- switch to state stCmdBytePrep after sequence 2 
      -- if there are more configuration bytes to send, keeps the state
         -- during sequences 3 and 4,
      -- if the configuration string was completely sent, switch to  
         -- state stDataBytePrep, to send the text message.
   -- While sending the text message, the Lcd_Mess_Ctrl   state machine:
      -- keeps state stDataBytePrep during sequences 4 and 5, 
      -- switch to state stDataByteReq after sequence 0 and keeps the 
         -- state during sequence 1, 
      -- switch to state stDataBytePrep after sequence 2 
      -- if there are more text bytes to send, keeps the state during 
         -- sequences 3 and 4,
      -- if the text string was completely sent, switch  to state 
         -- stMessIdle.

------------------------------------------------------------------------
-- Revision History:
--   12/21/2004(MirceaD): created
--   03/24/2005(MirceaD): formatted
------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
library work;
use work.LCD_pkg.all;

entity LCD_mess_Ctrl is
   Port (
-- General purpose signals
      ck: in std_logic;                   -- 3.125MHz clock
      reset: in std_logic;                -- btn1 on the main board 
-- User application interface signals
      lcd_config_string: in config_string_type;   
             -- the configuration string to be sent to the LCD
      LCD_message:in string;  -- the text string to be sent to the LCD
      LCD_mes_req: in std_logic; 
             -- the user application request to send a message
      LCD_mes_ack: out std_logic;                  
             -- the acknowledge signal generated by the Lcd_Mess_Ctrl
-- dbIOxCtrl component interface signals 
      lcd_RS_req: out std_logic;-- LCD Register Select request('1'-data,
                                --                    '0' - instruction)
      lcd_RW_req: out std_logic;-- LCD Read/Write request ('1' - read, 
                                --                         '0' - write)
      LCD_write_byte: out std_logic_vector (7 downto 0);   
             -- the byte to send to LCD
      LCD_req: out std_logic;             -- LCD cycle request
      LCD_ack: in std_logic               -- LCD cycle acknowledge
      );                        

end LCD_mess_Ctrl;

architecture Behavioral of LCD_mess_Ctrl is

------------------------------------------------------------------------
-- Constant and Signal Declarations
------------------------------------------------------------------------

-- The following constants define state codes for the LCD_mess 
-- controller state machine. The states are such a way assigned to
-- directly encode the output signals:
-- Lcd_Mes_Ack, Lcd_req, Cfg_Done
-- That avoids using output logic and generates glitch free output  
-- signals. The states are such a way assigned that each transition
-- changes a single state register bit (Grey code - like)
   constant stMessIdle     : std_logic_vector(2 downto 0) := "000";  
   constant stCmdBytePrep  : std_logic_vector(2 downto 0) := "100";   
   constant stCmdByteReq   : std_logic_vector(2 downto 0) := "110";   
   constant stDataBytePrep : std_logic_vector(2 downto 0) := "101";
   constant stDataByteReq  : std_logic_vector(2 downto 0) := "111";

-- LcdMess state register and next state 
   signal stLcdMessCur   : std_logic_vector(2 downto 0) := stMessIdle;

-- The attribute lines below prevent the ISE compiler to extract and 
-- optimize the state machines.
-- WebPack 5.1 doesn't need them (the default value is NO)
-- WebPack 6.2 has the default value YES, so without these lines would 
-- "optimize" the state machines.
-- Although the overall circuit would be optimized, the particular goal 
-- of "glitch free output signals" may not be reached. 
-- That is the reason of implementing the state machine as described in  
-- the constant declarations above. 

   attribute fsm_extract : string;
   attribute fsm_extract of stLcdMessCur: signal is "no"; 

   attribute fsm_encoding : string;
   attribute fsm_encoding of stLcdMessCur: signal is "user"; 

   attribute signal_encoding : string;
   attribute signal_encoding of stLcdMessCur: signal is "user"; 

   constant LCD_index_max : integer := 41;
   constant LCD_init_index_max : integer := 7;

   signal LCD_index : integer range 1 to LCD_index_max;

------------------------------------------------------------------------
-- Function Declaration
------------------------------------------------------------------------
function conv_ASCII(char:character) return STD_LOGIC_VECTOR is
-- Returns the ASCII code of the character argument
   variable ASCII:std_logic_vector(7 downto 0);
begin
   if char=' ' then ASCII:=X"20";
   elsif char='!' then ASCII:=X"21";
--   elsif char='"' then ASCII:=X"22";   -- forbidden character
   elsif char='#' then ASCII:=X"23";
   elsif char='$' then ASCII:=X"24";
   elsif char='%' then ASCII:=X"25";
   elsif char='&' then ASCII:=X"26";
   elsif char=''' then ASCII:=X"27";
   elsif char='(' then ASCII:=X"28";
   elsif char=')' then ASCII:=X"29";
   elsif char='*' then ASCII:=X"2A";
   elsif char='+' then ASCII:=X"2B";
   elsif char=',' then ASCII:=X"2C";
   elsif char='-' then ASCII:=X"2D";
   elsif char='.' then ASCII:=X"2E";
   elsif char='/' then ASCII:=X"2F";
   elsif char='0' then ASCII:=X"30";
   elsif char='1' then ASCII:=X"31";
   elsif char='2' then ASCII:=X"32";
   elsif char='3' then ASCII:=X"33";
   elsif char='4' then ASCII:=X"34";
   elsif char='5' then ASCII:=X"35";
   elsif char='6' then ASCII:=X"36";
   elsif char='7' then ASCII:=X"37";
   elsif char='8' then ASCII:=X"38";
   elsif char='9' then ASCII:=X"39";
   elsif char=':' then ASCII:=X"3A";
   elsif char=';' then ASCII:=X"3B";
   elsif char='<' then ASCII:=X"3C";
   elsif char='=' then ASCII:=X"3D";
   elsif char='>' then ASCII:=X"3E";
   elsif char='?' then ASCII:=X"3F";
   elsif char='@' then ASCII:=X"40";
   elsif char='A' then ASCII:=X"41";
   elsif char='B' then ASCII:=X"42";
   elsif char='C' then ASCII:=X"43";
   elsif char='D' then ASCII:=X"44";
   elsif char='E' then ASCII:=X"45";
   elsif char='F' then ASCII:=X"46";
   elsif char='G' then ASCII:=X"47";
   elsif char='H' then ASCII:=X"48";
   elsif char='I' then ASCII:=X"49";
   elsif char='J' then ASCII:=X"4A";
   elsif char='K' then ASCII:=X"4B";
   elsif char='L' then ASCII:=X"4C";
   elsif char='M' then ASCII:=X"4D";
   elsif char='N' then ASCII:=X"4E";
   elsif char='O' then ASCII:=X"4F";
   elsif char='P' then ASCII:=X"50";
   elsif char='Q' then ASCII:=X"51";
   elsif char='R' then ASCII:=X"52";
   elsif char='S' then ASCII:=X"53";
   elsif char='T' then ASCII:=X"54";
   elsif char='U' then ASCII:=X"55";
   elsif char='V' then ASCII:=X"56";
   elsif char='W' then ASCII:=X"57";
   elsif char='X' then ASCII:=X"58";
   elsif char='Y' then ASCII:=X"59";
   elsif char='Z' then ASCII:=X"5A";
   elsif char='[' then ASCII:=X"5B";
   elsif char='\' then ASCII:=X"5C";
   elsif char=']' then ASCII:=X"5D";
   elsif char='^' then ASCII:=X"5E";
   elsif char='_' then ASCII:=X"5F";
   elsif char='`' then ASCII:=X"60";
   elsif char='a' then ASCII:=X"61";
   elsif char='b' then ASCII:=X"62";
   elsif char='c' then ASCII:=X"63";
   elsif char='d' then ASCII:=X"64";
   elsif char='e' then ASCII:=X"65";
   elsif char='f' then ASCII:=X"66";
   elsif char='g' then ASCII:=X"67";
   elsif char='h' then ASCII:=X"68";
   elsif char='i' then ASCII:=X"69";
   elsif char='j' then ASCII:=X"6A";
   elsif char='k' then ASCII:=X"6B";
   elsif char='l' then ASCII:=X"6C";
   elsif char='m' then ASCII:=X"6D";
   elsif char='n' then ASCII:=X"6E";
   elsif char='o' then ASCII:=X"6F";
   elsif char='p' then ASCII:=X"70";
   elsif char='q' then ASCII:=X"71";
   elsif char='r' then ASCII:=X"72";
   elsif char='s' then ASCII:=X"73";
   elsif char='t' then ASCII:=X"74";
   elsif char='u' then ASCII:=X"75";
   elsif char='v' then ASCII:=X"76";
   elsif char='w' then ASCII:=X"77";
   elsif char='x' then ASCII:=X"78";
   elsif char='y' then ASCII:=X"79";
   elsif char='z' then ASCII:=X"7A";
   elsif char='{' then ASCII:=X"7B";
-- elsif char='|' then ASCII:=X"7C";  
      -- ASCII := x"00", this character is used as text message end mark
   elsif char='}' then ASCII:=X"7D";
   elsif char='~' then ASCII:=X"7E";
-- elsif char='<-' then ASCII:=X"7F"; -- non printable character
   else ASCII:=X"00";
   end if;
   return ASCII;
end;

begin
------------------------------------------------------------------------
-- Module Implementation
------------------------------------------------------------------------
------------------------------------------------------------------------
-- LCD Message Controller State Machine
------------------------------------------------------------------------

Lcd_Mess_Ctrl_FSM:   process (ck, reset)
begin
   if reset = '1' then
      stLcdMessCur <= stMessIdle; 
   elsif ck'Event and ck = '1' then
      case stLcdMessCur is    
         when stMessIdle =>
         -- Idle state waiting for the beginning of an LCD Message cycle
            if Lcd_Mes_req = '1' then -- there is a new message for LCD
               LCD_index <= 1;        -- first index prepared
               stLcdMessCur <= stCmdBytePrep;
--          else                      
--             stLcdMessCur <= stMessIdle;   -- Keep old state (default)
            end if;

         when stCmdBytePrep =>   -- Prepare to send a Command byte
            if Lcd_Ack = '0' then-- dbIOxCtrl is ready to receive a byte
               if LCD_config_string(LCD_index)(7 downto 0)/=x"00" then 
                     -- the command string is not over
                  LCD_write_byte <= 
                               LCD_config_string(LCD_index)(7 downto 0);
                     -- a new byte is sent to LCD
                  LCD_RS_req <= '0';         -- it is a command byte
                  LCD_RW_req <= '0';         -- write cycle
                  LCD_index <= LCD_index + 1;
                     -- next index prepared (for next time)
                     stLcdMessCur <= stCmdByteReq;
               else
                  LCD_index <= 1;            -- first index prepared
                     stLcdMessCur <= stDataBytePrep;
                  end if;
--          else
--             stLcdMessCur <= stCmdBytePrep;-- Keep old state (default)
            end if;

         when stCmdByteReq =>      -- Request to send a Command byte
            if   Lcd_Ack = '1' then   -- dbIOxCtrl received the byte
               stLcdMessCur <= stCmdBytePrep;
--          else
--             stLcdMessCur <= stCmdByteReq; -- Keep old state (default)
            end if;
  
         when stDataBytePrep =>  -- Prepare to send a Data (text) byte
            if Lcd_Ack = '0' then-- dbIOxCtrl is ready to receive a byte
               if conv_ASCII(LCD_message(LCD_index)) /= x"00" then 
                      -- the data string is not over
                  LCD_write_byte <= conv_ASCII(LCD_message(LCD_index));
                      -- a new byte is sent to LCD
                  LCD_RS_req <= '1';               -- it is a data byte
                  LCD_RW_req <= '0';               -- write cycle
                  LCD_index <= LCD_index + 1;      
                      -- next index prepared (for next time)
                     stLcdMessCur <= stDataByteReq;
               elsif   Lcd_Mes_Req = '0' then
                  stLcdMessCur <= stMessIdle;
--             else
--                stLcdMessCur <= stDataBytePrep;-- Keep old state (def)
               end if;
--          else
--             stLcdMessCur <= stDataBytePrep;- Keep old state (default)
            end if;

         when stDataByteReq =>   -- Request to send a Data (text) byte
            if Lcd_Ack = '1' then   -- dbIOxCtrl received the byte
               stLcdMessCur <= stDataBytePrep;
--          else
--             stLcdMessCur <= stDataByteReq;-- Keep old state (default)
            end if;

         when others =>         -- Unknown states
            stLcdMessCur <= stMessIdle;

      end case;
   end if;
end process;

------------------------------------------------------------------------
-- Output signals assignment (state bits)
------------------------------------------------------------------------

Lcd_Mes_Ack <= stLcdMessCur(2);
Lcd_req       <= stLcdMessCur(1);

end Behavioral;
