------------------------------------------------------------------------
--  ck_div.vhd -- Clock divider
------------------------------------------------------------------------
--  Author : Mircea Dabacan
--           Copyright 2004 Digilent, Inc.
------------------------------------------------------------------------
-- Software version: Xilinx ISE 6.2.03i
--                   WebPack
------------------------------------------------------------------------
-- This is the source file for the ck_div component, 
-- used in the "Tennis" Digilent Reference Project.

------------------------------------------------------------------------
--  Port definitions
------------------------------------------------------------------------
-- ck   : in std_logic;      -- 50MHz board clock
-- ck3MHz: out std_logic;    -- clock for dbIOxCtrl, Lcd_Mess_Ctrl
-- ck1KHz: out std_logic;    -- clock for DIO2 or DIO5 CPLD chip 
                             -- (7 segment display)
-- ck24Hz: out std_logic);   -- clock for Pabi_Waledi_FSM

------------------------------------------------------------------------
--  Internal signals
------------------------------------------------------------------------
-- cnt: std_logic_vector(20 downto 0);  
        -- ripple carry counter used as frequency divider

------------------------------------------------------------------------
--  Behavioral description
------------------------------------------------------------------------
-- The file describes a ripple carry counter. 
-- The dbIO2Ctrl and Lcd_Mess_Ctrl components use same clock (ck3MHz),
-- so they are any way synchronized (a BUFG is instantiated for ck3MHz). 
-- Pabi_Waledi FSM component uses a much slower clock. 
-- The protocols used at component interfaces allow asynchronous clocks.
-- From these reasons the clock propagation delays are not critical.
-- They allow using a ripple carry counter as clock divider. It is much 
-- simpler and consumes much fewer resources than a synchronous one.

------------------------------------------------------------------------
-- Revision History:
--  12/21/2004(MirceaD): created
--  03/24/2005(MirceaD): formatted
------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity ck_div is
  Port ( ck : in std_logic;    -- 50MHz board clock
    ck3MHz : out std_logic;    -- clock for dbIOxCtrl, Lcd_Meaa_Ctrl
    ck1KHz : out std_logic;    -- clock for DIO2 or DIO5 CPLD chip 
                               -- (7 segment display)
    ck24Hz : out std_logic);   -- clock for Pabi_WalediFSM
end ck_div;

architecture behavioral of ck_div is

signal cnt: std_logic_vector(20 downto 0);  
       -- ripple carry counter used as frequency divider

component BUFG    -- bufg declared for ck3MHz
   port (I: in std_logic; O: out std_logic);
end component;   

begin

divider: process (ck)  -- ripple carry counter. 
-- Successive clock signals are allowed to have different phase shift.
begin

  if ck'event and ck = '1' then
      cnt(0) <= not cnt(0);    -- divide by 2^16 = 65536 
      -- (not critical to have precisely 1 KHz frequency)
  end if;

  for i in 1 to 20 loop
    if cnt(i-1)'event and cnt(i-1)='1' then
      cnt(i) <=  not cnt(i);
    end if;
  end loop;

end process;

  ck24Hz <= cnt(20);  -- no bufg. 
  -- ck24Hz is further divided (by 5 or 8) in the PAbi_Waledi_FSM module
  ck1KHz <= cnt(15);  -- no bufg. 
  -- ck1KHz is directly connected to the output pin. No jitter problems.

  ck3MHz_buf: BUFG port map (I => cnt(3), O => ck3MHz);
  -- bufg instantiated ck3MHz

end behavioral;
